package truebar.example;

import vitasis.truebar.client.AudioRecorder;
import vitasis.truebar.client.TruebarAuthenticator;
import vitasis.truebar.client.TruebarClient;
import vitasis.truebar.client.misc.Pipeline;
import vitasis.truebar.client.model.Metadata;
import vitasis.truebar.client.model.ServiceParameters;
import vitasis.truebar.client.model.pipeline.*;
import vitasis.truebar.client.model.pipeline.config.StageConfigIn;

import javax.sound.sampled.*;
import java.nio.file.Path;
import java.util.*;
import java.util.concurrent.TimeUnit;
import javax.sound.sampled.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;


/* *****************************************************
    !IMPORTANT!
    This file is just for demonstration purposes.
    TruebarClient can be used asynchronously without blocking main thread instead of waiting for responses like it is done in this example.
 ******************************************************** */

public class Main {

    private final static String clientCode = "JAVA_EXAMPLE";

    public static void main(String[] args) {
        try {
            Scanner scanner = new Scanner(System.in);

            // ************************************
            // Authentication
            // ************************************
            System.out.print("Enter username: ");
            String userName = scanner.nextLine();
            System.out.print("Enter password: ");
            String password = scanner.nextLine();

            // Create truebar authenticator used to manage authentication details
            TruebarAuthenticator truebarAuthenticator = new TruebarAuthenticator(
                    new ServiceParameters(
                            true,
                            "dev-pipelines-auth.true-bar.si",
                            443,
                            3000
                    ),
                    userName,
                    password
            );

            // Create truebar client instance
            TruebarClient truebarClient = new TruebarClient(
                    truebarAuthenticator,
                    new ServiceParameters(
                            true,
                            "dev-api-pipelines.true-bar.si",
                            443,
                            180000
                    )
            );

            // Select scenario
            System.out.print("Select scenario\n1 - STT\n2 - TTS\n3 - STS: ");
            int scenarioSelection = Integer.parseInt(scanner.nextLine());

            if (scenarioSelection == 1) {
                // ************************************
                // Create pipeline using 2 stages: "automatic speech recognition" and "automatic punctuation"
                // Each stage has multiple config options and each config option provides different set of parameters.
                //************************************

                // List all config option provided by ASR stage and print selection menu
                StageIn asrStageIn = truebarClient.getStage(Task.ASR).get();
                List<StageConfigIn> asrStageOptions = asrStageIn.configOptions();
                System.out.println("Available ASR options:");
                for (int i = 0; i < asrStageOptions.size(); i++) {
                    System.out.printf("%d -> %s%n", i + 1, asrStageOptions.get(i).tag());
                }
                System.out.print("Enter selection: ");
                StageConfigIn selectedAsrConfigOption = asrStageOptions.get(Integer.parseInt(scanner.nextLine()) - 1);

                // List all config option provided by PC stage and print selection menu
                StageIn pcStageIn = truebarClient.getStage(Task.NLP_pc).get();
                List<StageConfigIn> pcStageOptions = pcStageIn.configOptions();
                System.out.println("Available punctuation options:");
                System.out.printf("%d -> %s%n", 0, "None");
                for (int i = 0; i < pcStageOptions.size(); i++) {
                    System.out.printf("%d -> %s%n", i + 1, pcStageOptions.get(i).tag());
                }
                System.out.print("Enter selection: ");
                StageConfigIn selectedPcConfigOption = null;
                int selectedPcIndex = Integer.parseInt(scanner.nextLine());
                if (0 < selectedPcIndex) {
                    selectedPcConfigOption = pcStageOptions.get(selectedPcIndex - 1);
                }

                // Create pipeline with ASR -> PC stages
                Pipeline pipeline = new Pipeline();

                // ASR stage - enable interim transcripts on ASR stage if supported by selected config option
                Map<String, Object> asrStageParameters = new HashMap<>();
                if (selectedAsrConfigOption.parameters().containsKey("enableInterims")) {
                    asrStageParameters.put("enableInterims", true);
                }
                pipeline.addStage(Task.ASR, selectedAsrConfigOption.tag(), StageOut.ExceptionHandlingPolicy.THROW, asrStageParameters);

                // PC stage - Add pc stage if selected (without any specific parameters)
                if (selectedPcConfigOption != null) {
                    pipeline.addStage(Task.NLP_pc, selectedPcConfigOption.tag(), StageOut.ExceptionHandlingPolicy.RETRY_AND_THROW, Map.of());
                }

                System.out.print("Options:\n1 -> Start streaming session\n2 -> Process file\nEnter selection: ");
                switch (scanner.nextLine()) {
                    case "1": {
                        System.out.println("Creating session...");
                        TruebarClient.StreamingSession streamingSession = truebarClient.process(
                                        pipeline,
                                        (textSegment) -> System.out.println(createHypothesis(textSegment)),
                                        (warning) -> System.err.printf("Received following warning from truebar service: %s%n", warning),
                                        (error) -> System.err.printf("Received following error from truebar service: %s%n", error.getTruebarErrorIn().message()),
                                        new Metadata(clientCode, null),
                                        null
                                )
                                .get(300, TimeUnit.SECONDS);

                        System.out.println("Available audio inputs:");
                        List<Mixer.Info> audioSources = AudioRecorder.findAvailableAudioMixers();
                        for (int i = 0; i < audioSources.size(); i++) {
                            System.out.printf("%d -> %s%n", i + 1, audioSources.get(i).getName());
                        }
                        Mixer.Info audioInputInfo = audioSources.get(Integer.parseInt(scanner.nextLine()) - 1);

                        try (AudioRecorder audioRecorder = new AudioRecorder(audioInputInfo)) {
                            System.out.println("...CAPTURING AUDIO .....TO FINISH PRESS ENTER...");
                            audioRecorder.start((audioChunk) -> streamingSession.sendData(audioChunk.array()));

                            // Loop until user presses enter
                            while (System.in.available() == 0) {
                                Thread.sleep(100);
                            }

                            System.out.println("FINISHING SESSION");
                            audioRecorder.stop();
                            streamingSession.finish(false).get(3, TimeUnit.SECONDS);
                        }
                        break;
                    }
                    case "2": {
                        System.out.print("Enter path to audio file: ");
                        String audioFilePath = scanner.nextLine();

                        System.out.println("Uploading file...");
                        List<TextSegmentDto> result = truebarClient
                                .process(pipeline, Path.of(audioFilePath), new Metadata(clientCode, null))
                                .get();

                        System.out.println("Result:");
                        for (TextSegmentDto textSegment : result) {
                            System.out.println(createHypothesis(textSegment));
                        }
                        break;
                    }
                    default: {
                        throw new RuntimeException("Invalid selection!");
                    }
                }


            }
            // !!!!!!!!!!!!!!!! TTS !!!!!!!!!!!!!!!!!!!!!!!
            else if (scenarioSelection == 2) {
                // List all config option provided by TTS stage and print selection menu
                StageIn ttsStageIn = truebarClient.getStage(Task.TTS).get();
                List<StageConfigIn> ttsStageOptions = ttsStageIn.configOptions();
                System.out.println("Available TTS options:");
                for (int i = 0; i < ttsStageOptions.size(); i++) {
                    System.out.printf("%d -> %s%n", i + 1, ttsStageOptions.get(i).tag());
                }
                System.out.print("Enter selection: ");
                StageConfigIn selectedTtsConfigOption = ttsStageOptions.get(Integer.parseInt(scanner.nextLine()) - 1);

                // Create pipeline containing just selected TTS stage
                Pipeline pipeline = new Pipeline();
                pipeline.addStage(Task.TTS, selectedTtsConfigOption.tag(), StageOut.ExceptionHandlingPolicy.SKIP, Map.of());


                System.out.print("Options:\n1 -> Start streaming session\n2 -> Process sentences\nEnter selection: ");
                switch (scanner.nextLine()) {
                    case "1": {
                        throw new RuntimeException("Not implemented!");
                    }

                    case "2": {
                        StreamingAudioPlayer streamingAudioPlayer = new StreamingAudioPlayer(new AudioFormat(16000.0f, 16, 1, true, false));
                        streamingAudioPlayer.start();

                        while (true) {
                            System.out.print("Enter sentence to systematize or leave blank to exit: ");
                            String line = scanner.nextLine();
                            if (line == null || line.isEmpty()) {
                                break;
                            }

                            byte[] response = truebarClient.processTts(
                                    pipeline,
                                    List.of(new TextSegmentDto(
                                            true,
                                            null,
                                            null,
                                            Arrays
                                                    .stream(line.split(" "))
                                                    .map(w -> new TokenDto(
                                                            null,
                                                            null,
                                                            false,
                                                            false,
                                                            w,
                                                            null
                                                    ))
                                                    .toList()
                                    )),
                                    new Metadata("JAVA_EXAMPLE", null)
                            ).get();


                            streamingAudioPlayer.enqueue(response);
                        }

                    }

                    default: {
                        throw new RuntimeException("Invalid selection!");
                    }
                }
            } else if (scenarioSelection == 3) {
                StageIn asrStageIn = truebarClient.getStage(Task.ASR).get();
                List<StageConfigIn> asrStageOptions = asrStageIn.configOptions();
                System.out.println("Available ASR options:");
                for (int i = 0; i < asrStageOptions.size(); i++) {
                    System.out.printf("%d -> %s%n", i + 1, asrStageOptions.get(i).tag());
                }
                System.out.print("Enter selection: ");
                StageConfigIn selectedAsrConfigOption = asrStageOptions.get(Integer.parseInt(scanner.nextLine()) - 1);

                // List all config option provided by NLP_nmt stage and print selection menu
                StageIn nmtStageIn = truebarClient.getStage(Task.NLP_nmt).get();
                List<StageConfigIn> nmtStageOptions = nmtStageIn
                        .configOptions()
                        .stream()
                        .filter(configOption -> selectedAsrConfigOption.tag().getLanguage().equals(configOption.tag().getLanguage().split(",")[0]))
                        .toList();
                System.out.println("Available NMT options:");
                for (int i = 0; i < nmtStageOptions.size(); i++) {
                    System.out.printf("%d -> %s%n", i + 1, nmtStageOptions.get(i).tag());
                }
                System.out.print("Enter selection: ");
                StageConfigIn selectedNmtConfigOption = nmtStageOptions.get(Integer.parseInt(scanner.nextLine()) - 1);

                // List all config option provided by TTS stage and print selection menu
                StageIn ttsStageIn = truebarClient.getStage(Task.TTS).get();
                List<StageConfigIn> ttsStageOptions = ttsStageIn
                        .configOptions()
                        .stream()
                        .filter(configOption -> selectedNmtConfigOption.tag().getLanguage().split(",")[1].equals(configOption.tag().getLanguage()))
                        .toList();
                System.out.println("Available TTS options:");
                for (int i = 0; i < ttsStageOptions.size(); i++) {
                    System.out.printf("%d -> %s%n", i + 1, ttsStageOptions.get(i).tag());
                }
                System.out.print("Enter selection: ");
                StageConfigIn selectedTtsConfigOption = ttsStageOptions.get(Integer.parseInt(scanner.nextLine()) - 1);

                // Create pipeline consisting of ASRm NMT and TTS stages
                Pipeline pipeline = new Pipeline();
                pipeline.addStage(Task.ASR, selectedAsrConfigOption.tag(), StageOut.ExceptionHandlingPolicy.THROW, Map.of());
                pipeline.addStage(Task.NLP_nmt, selectedNmtConfigOption.tag(), StageOut.ExceptionHandlingPolicy.THROW, Map.of());
                pipeline.addStage(Task.TTS, selectedTtsConfigOption.tag(), StageOut.ExceptionHandlingPolicy.THROW, Map.of());

                // Create audio player that will be used to play received audio.
                StreamingAudioPlayer streamingAudioPlayer = new StreamingAudioPlayer(new AudioFormat(16000.0f, 16, 1, true, false));
                streamingAudioPlayer.start();

                System.out.println("Creating session...");
                TruebarClient.StreamingSession streamingSession = truebarClient.process(
                                pipeline,
                                streamingAudioPlayer::enqueue,
                                (warning) -> System.err.printf("Received following warning from truebar service: %s%n", warning),
                                (error) -> System.err.printf("Received following error from truebar service: %s%n", error.getTruebarErrorIn().message()),
                                new Metadata(clientCode, null)
                        )
                        .get(300, TimeUnit.SECONDS);

                System.out.println("Available audio inputs:");
                List<Mixer.Info> audioSources = AudioRecorder.findAvailableAudioMixers();
                for (int i = 0; i < audioSources.size(); i++) {
                    System.out.printf("%d -> %s%n", i + 1, audioSources.get(i).getName());
                }
                Mixer.Info audioInputInfo = audioSources.get(Integer.parseInt(scanner.nextLine()) - 1);

                // Start audio recorder. Manages audio capture thread internally
                try (AudioRecorder audioRecorder = new AudioRecorder(audioInputInfo)) {
                    System.out.println("...CAPTURING AUDIO .....TO FINISH PRESS ENTER...");
                    audioRecorder.start((audioChunk) -> streamingSession.sendData(audioChunk.array()));

                    // Loop until user presses enter
                    while (System.in.available() == 0) {
                        Thread.sleep(100);
                    }

                    System.out.println("FINISHING SESSION");
                    audioRecorder.stop();
                    streamingSession.finish(false).get(3, TimeUnit.SECONDS);
                }

                streamingAudioPlayer.stop();
            } else {
                throw new RuntimeException("Invalid scenario selection!");
            }

            System.out.println("ALL OK");
        } catch (Exception e) {
            System.out.println("Caught unhandled exception.");
            e.printStackTrace();
        }
    }

    private static String createHypothesis(TextSegmentDto textSegment) {
        StringBuilder hypothesis = new StringBuilder();
        boolean previousRightHanded = false;
        for (TokenDto tokenDto : textSegment.tokens()) {
            if (!previousRightHanded && !tokenDto.isLeftHanded()) {
                hypothesis.append(" ");
            }
            hypothesis.append(tokenDto.text());
            previousRightHanded = tokenDto.isRightHanded();
        }
        return hypothesis.toString();
    }

    public static class StreamingAudioPlayer {
        private final BlockingQueue<byte[]> audioQueue = new LinkedBlockingQueue<>();
        private SourceDataLine sourceLine;
        private final AudioFormat audioFormat;
        private volatile boolean playing = true;

        public StreamingAudioPlayer(AudioFormat format) {
            this.audioFormat = format;
        }

        public void start() throws LineUnavailableException {
            DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioFormat);
            sourceLine = (SourceDataLine) AudioSystem.getLine(info);
            sourceLine.open(audioFormat);
            sourceLine.start();

            // Playback thread
            new Thread(() -> {
                try {
                    while (playing) {
                        byte[] chunk = audioQueue.take(); // blocks if no data
                        sourceLine.write(chunk, 0, chunk.length);
                    }
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            }, "AudioPlayerThread").start();
        }

        public void stop() {
            playing = false;
            sourceLine.drain();
            sourceLine.stop();
            sourceLine.close();
        }

        public void enqueue(byte[] audioChunk) {
            try {
                audioQueue.put(audioChunk);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                throw new RuntimeException("Interrupted while waiting to put data in audio input queue.", e);
            }
        }
    }
}